'
' locale.bi
'
' FreeBASIC localisation library
'
' provides ample convenient localisation support
'
' STR files are stringtables, UTF-8 encoded in a special "INI" format:
' use ' or ! for comment lines
' <String-ID> "<String>" - string, addressed by "String-ID", which must be a
'     word in latin letters, case insensitive, valid chars: a-z A-Z and 0-9, further "_"
'     space chars (code<33) will be ignored, a line starting with other chars will be ignored
'     you may put a "=" or ":" between ID and string, everything is ignored from the first
'     illegal character to the first quotation marks. The returned string will be everything
'     between the quotation marks. The string is UTF-8 encoded and may contain any character
'     except codes below 32 or at least not 13, 10 and 0. You can use '\r' for linebreaks(13),
'     '\e' for code 27 (Escape), '\t' for tabulator(8), '\"' to mask quotation marks and
'     '\\' for backslashs. Any single '\' will disappear.
' <String-ID>
'   <Text>
' END
'     => Multiline string entry, must not contain comments, the whole text between string-ID
'     and "END" counts as string, line breaks are turning into space chars, use \r instead.
' [<Section>] starts a new section. You may address strings by "<sectionname>.<string ID>"
'     where "sectionname" may also contain one or more "." for imaginary subsections.
'     The section expression might be also a language code where the search for strings
'     for a string in a distinct language will start resp. end. If you name a section "[english]"
'     it will not be recognized as English string table block, you get a section named "english".
'     Section names should be at least 6 characters long and must not equal an language code.
'     A language code consists of the ISO 639-1 language code and, divided by an underscore,
'     a region following ISO-3166, for example "en_US" for US-English, "el_GR" is Greek (Greece)
'     Each language can only have one block in one file, only the first found is valid. Only for use
'     in multilingual STR files, i.e. files named e.g. "en.str" or being in a folder named "en"
'     are assumed to contain strings of the language "en"
'     NOTE: unlike string IDs, the section code is only valid if standing at the line start
'     with no space chars
'     NOTE: ISO-639-3 (3-letter code) and ISO-3166-2 support will follow
'
' basic.str: the basic string table, loaded first. It contains among other things all language labels
'     IDS_LANGLABEL contains the localised string for a language, e.g. "Deutsch" for German in German,
'     only upper case if strictly upper case.
' system.str: interface strings, especially GUI. Some are button labels, therefore upper case, and
'     containing hotkey codes. Most strings are for a very special purpose and some would exist
'     seemingly double. But in some languages they are different. It's like merging a word
'     in a sentence string - it won't always lead to a correct translation.
'
' You can request a string by its ID and get the version for the specified language. The
' language code consists of language+country like "en_US" for English / United States Of America
' If you request a string, first a string for language+country is searched, then generic
' language, and at last the English - "en" version gets returned. Inexistant IDs will return
' an error message "!ERR_NOT_FOUND..." - you may check just for a leading "!ERR" or simply "!"
' in the returned string
'
' There's only one global namespace, the later loaded files are searched first, so you might
' also "overwrite" strings with the same ID. Don't use string IDs starting with "IDS_", those
' a reserved for "system" functions.
'
' #todo: library expects a two-letter language code, should also support ISO-639-3 codes, resp.
'        switch, add translation for downward compatibility
'
' LOCALESETLANGUAGE(STRING newlanguage) sets language to use by the library - not the user's
'                 language, reloads all string table files
'
' LOCALEGETLANGUAGE() returns the currently set language of the library, like set by
'                 LOCALESETLANGUAGE(); does not return the system's or users's language
'
' LOCALEGETSTRING(STRING id) returns a localised string by its ID
'                 first searches the explicitely loaded language, then the English fallback
'                 all illegal characters like "%" in the "id" string will be removed before
'                 the search
'
' LOCALEGETLANGUAGELABEL(STRING language) returns the localised name/"label" of specified
'                 language written in its language
'                 the labels are stored in the basic string table, all other strings can
'                 only be requested in the loaded language and English(fallback)
'
' LOCALEGETLANGUAGEPACK(STRING file_id="") returns available languages of given STR file
'                 like DIR(): after the first call you'll get language codes
'                 one by one just calling "LOCALEGETLANGUAGEPACK()"
'                 You can get a matching paraphrase by calling
'                 LOCALEGETSTRING("IDS_LANGUAGE", <language code>)
'                 - regard to convert the first letter to upper case (with UTF-8 function)
'                 file_id might be a path ".../locale/" or a similar file ".../locale/x.str"
'                 where "x.str" might be searched for language sections, otherwise
'                 it depends on folder and file names (e.g. "en/x.str" / "en_US.str" )
'                 => 2-letter-filenames or 2x2 with underscore between
'
'
' LOCALENEW([STRING language]) initialises localisation library, loads basic strings among other things
'
' LOCALELOADTABLE(STRING filename) searches&loads given STR file,
'                          use e.g. "mystrings.str", path+filename or path without filename
'                          ending with a "/"
'                          if the specified file exists directly, - e.g. ".../locale/mystr.str" -
'                          it is assumed, that it contains the strings for all languages.
'                          otherwise the files named "mystr.str" will be loaded from the
'                          locations "...locale/de/de_DE/", "...locale/de/" and "...locale/en/"
'                          if the language code is "de_DE"
'                          if no filename is specified, "...locale/de_DE.str", "...locale/de.str"
'                          and "...locale/en.str" are tried to be loaded
'
' LOCALUNLOADTABLE(STRING filename) unloads given STR file
'
' LOCALEDATE(LONGINT Unix time[, STRING format]) - turns a given Unix time/Universal time value into
' (LOCALTIME(...))         a string formulation, "ALIAS" LOCALETIME - LOCALDATE creates at default
'                          a short date formulation like "<year>/<month>/<day>", while LOCALTIME
'                          creates a short time formulation like "<24 hour>:<minute>"
'                          #note: there is no timezone offset, you have to convert a real Universal time
'                          value into local time by adding or substracting hours (you may also store
'                          a local time value in Unix format)
'                  format: UTF-8 string, containing value keys translating the time value into
'                          corresponding substrings
'                          %s "0"-"59" second, %ss "00-59" second
'                          %m 0-59 minute, %mm 00-59 minute
'                          %h 1-12 hour, %hh 01-12 hour, %H 0-23 hour, %HH 00-23 hour (%_h,%_H) %ap "AM"/"PM"
'                          %D 1-31 day, %DD 01-31 day (%_D)
'                          %Wd "Mo"-"Su" weekday, %Wdy "Mon"-"Sun"(3 chars => also "Mo."), %Wday "Monday"-"Sunday" (+%wday)
'                          %Dth ordinal day number "1st"-"31th" (of course no "01st", "02nd", but %_Dth)
'                          %M 1-12 month, %MM 01-12 (%_M), %Mn "Ja"-"De", %Mon "Jan"-"Dec", %Month "January"-"December"
'                          %Y 1-nnnn year, %YY 01-nn year, %YYYY -nnnn-nnnn year with year 0, %Year 1-nnnn BC/AD without 0
'                          %ba "BC"/"AD" (important for %Year)
'                          %_H is e.g. like %HH, but using a leading Space instead of a leading "0" in case
'                          not supported for years, minutes and seconds
'                          use \% to mask a percent sign, \\ and \" will be translated too but \ and " could be used without
'                          escaping. Everything that is not a valid key or escape sequence will appear in the formatted string
'                          #todo: support of different calendar systems, starting e.g. with "!JEWISH:". Will disable
'                                 the default keys resp. replace; %YYYY could be the year of a different calendar system though.
'                                 Keys might also contain UTF-8 characters, but must be unique.
'                                 Default is "!GREGORIAN:"
' LOCALENUMBER(double value[, INTEGER digits|STRING format[, STRING language]]) - returns a formatted number string
'                          mostly for correctly using point and comma, rounded to two decimal places
'                          with thousands separators, digits are count of decimal places
'                  format: # for decimal digit
'                          0 digit always filled with a number, _ for leading Spaces instead of 0s
'                          . for decimal point character: .# optional, .0 and ._ forced, no "." forces rounded integer
'                          , for thousands separators
'                          Space for Space as thousands separator
'                          \# \0 \_ \, \. \\ to mask characters / put out without handling as part of the number.
'                          *10^^ and E+00 are forcing exponential notation => also *10^0 "...E+00"
'                          No variations possible, uses locale multiplication sign.
'                          Exponents from -99 to (+)99. The basic number (###...) gets normalised when
'                          "*10^^" or "E+00" are appearing, the "E+00" format is automatically used on
'                          numbers outside the range 10^-12 to 10^12 (>12 digits)
'                          Default is "#,##0.##" (equals "#,###,###,###,##0.##")
'                          The minus sign is the first number digit => "_00": -3 -> "-03" 5 -> " 05",
'                          "000": -3 -> "-03", 5 -> "005"
'                          You can't limit the length of the integer part.
'                          Example: "##\.##.#" would produce for the value 123456789.4321 "1.23.45.67.89.4321"
'                          - this is not an IP address
'                          #todo: support of different number systems e.g. "!BASE16:", may use different codes
'
' LOCALEDELETE de-initialises localisation library, frees buffers

#INCLUDE ONCE "prj.bi"
#INCLUDE ONCE "utf8.bi"
#INCLUDE ONCE "system.bi"
#INCLUDE ONCE "datafolder.bi"
#INCLUDE ONCE "fbext.bi"

DIM SHARED AS STRING localecurlanguage
DIM SHARED AS STRING localestrfilename(), localestrfile()
DIM SHARED AS INTEGER localestrfilecountsize
DIM SHARED AS STRING localelanguagepacklist
DIM SHARED AS INTEGER localelanguagepackindex

DECLARE SUB LOCALENEW(BYVAL language AS STRING="")
DECLARE SUB LOCALELOADTABLE(AS STRING)

DECLARE FUNCTION LOCALEORDINALNUMBER(BYVAL value AS INTEGER, BYVAL language AS STRING="") AS STRING
DECLARE FUNCTION LOCALENUMBER OVERLOAD (BYVAL value AS DOUBLE, BYVAL digits AS INTEGER, BYVAL language AS STRING="") AS STRING
DECLARE FUNCTION LOCALENUMBER OVERLOAD (BYVAL value AS DOUBLE, BYVAL formatstr AS STRING="", BYVAL language AS STRING="") AS STRING

SUB LOCALESETLANGUAGE(BYVAL newlanguage AS STRING="")
  IF localecurlanguage="" THEN LOCALENEW ' forgot to initialise
  IF newlanguage>"" THEN
    ' #todo: maybe validation
  ELSE ' no parameter => assign user's language
    newlanguage=GETUSERLANGUAGE()
  ENDIF

  IF newlanguage=localecurlanguage THEN RETURN ' no change

  ' set new language
  localecurlanguage=newlanguage

  ' reload all involved string tables
  '  #todo: make it much cleaner than this
  LOCALENEW(newlanguage)

END SUB

FUNCTION LOCALEGETLANGUAGE() AS STRING
  IF localecurlanguage="" THEN LOCALENEW ' forgot to initialise

  RETURN localecurlanguage
END FUNCTION

FUNCTION LOCALEGETLOCALPATH(BYVAL filename AS STRING) AS STRING
  ' internal function, determines existing path to local files
  ' only use filename without path
  DIM AS STRING localfilepath=""

  ' #note: although the file found is e.g.
  '        "<path>/locale/fr/mystrings.str", the returned path has to be
  '        "<path>/locale/mystrings.str"
  '        it's only determining the path resp. if it's containing the STR file
  IF filename="" THEN
    ' determine path to a file named <language>.str
    localfilepath="%APPDATA%locale/"+localecurlanguage+".str"
    IF DFEXISTS(localfilepath) THEN RETURN "%APPDATA%locale/"
  ELSE
    ' look in appdata-path (resp. local folder)
    IF DFEXISTS("%APPDATA%") THEN
      ' ...rsc:locale/<filename>
      localfilepath="%APPDATA%locale/"+filename
      IF DFEXISTS(localfilepath) THEN RETURN localfilepath

      ' e.g. ...rsc:locale/de/de_DE/<filename> (regional language)
      localfilepath="%APPDATA%locale/"+MID(localecurlanguage,1,2)+"/"+localecurlanguage+"/"+filename
      IF DFEXISTS(localfilepath) THEN
        RETURN "%APPDATA%locale/"+filename
      ENDIF

      ' e.g. ...rsc:locale/de/<filename> (main language)
      localfilepath="%APPDATA%locale/"+MID(localecurlanguage,1,2)+"/"+filename
      IF DFEXISTS(localfilepath) THEN
        RETURN "%APPDATA%locale/"+filename
      ENDIF
    ENDIF
  ENDIF

  ' default / local resource folder
  localfilepath=PRJGETID()+".rsc:locale/"+filename
  RETURN localfilepath
END FUNCTION

SUB LOCALENEW(BYVAL language AS STRING="")
  DIM fn AS INTEGER
  DIM AS STRING filename

  IF language="" THEN
    localecurlanguage=GETUSERLANGUAGE()
  ELSE
    localecurlanguage=language
  ENDIF

  ' the first stringtables are "system" strings, common strings,
  ' especially the names of languages in their language
  ' file(0) contains all strings in all supported languages
  '         => locale/basic.str
  ' file(1) are common interface strings for the current language like "Yes" and "No"
  '         => locale/<language>/system.str
  ' "System string" IDs are starting with "IDS_" and may contain a "&" you may
  ' have to remove. This case is for standard-GUI-button labelling, for instance.
  REDIM localestrfilename(2)
  REDIM localestrfile(2)
  localestrfilecountsize=2

  LOCALELOADTABLE("%FBDATA%locale/basic.str")
  LOCALELOADTABLE("%FBDATA%locale/system.str")
  IF localestrfilename(0)="" THEN ' basic string tables not yet found
    LOCALELOADTABLE("%FBDATA%locale/basic.str")
    LOCALELOADTABLE("%FBDATA%locale/system.str")
  ENDIF

END SUB

SUB LOCALEDELETE
  ' deconstructor
  ' deleting all used strings/arrays, free RAM...
  ' #todo:
END SUB

SUB LOCALELOADTABLE(file_id AS STRING)
  DIM AS INTEGER index, pslashindex, l, i, j
  DIM AS STRING filename, lcode, fbuffer, sectionid, keyid
  DIM AS UBYTE c

  IF localecurlanguage="" THEN LOCALENEW ' forgot to initialise
  lcode=localecurlanguage

  index=localestrfilecountsize
  DO
    index-=1
  LOOP WHILE index AND file_id<>localestrfilename(index)

  IF file_id<>localestrfilename(index) THEN ' unknown, new file => add new
    ' find next free file
    FOR i=0 TO localestrfilecountsize-1
      IF localestrfilename(i)="" THEN
        EXIT FOR
      ENDIF
    NEXT

    IF i>=localestrfilecountsize THEN ' array full
      index=localestrfilecountsize
      localestrfilecountsize+=1
      REDIM PRESERVE localestrfilename(localestrfilecountsize)
      REDIM PRESERVE localestrfile(localestrfilecountsize)
      localestrfilename(index)=file_id
    ELSE
      index=i
      localestrfilename(index)=file_id
    ENDIF
  ENDIF

  ' the form of file_id is "...locale/myfile.str"
  ' if this file exists, it is assumed that it contains all strings in all languages
  ' otherwise it'll be tried to load "...locale/<lang>/<lang_region>/myfile.str",
  ' "...locale/<lang>/myfile.str" and "...locale/en/myfile.str"
  ' if file_id is a path without filename, "...locale/<lang_region>.str",
  ' "...locale/<lang>.str" and "...locale/en.str"
  filename=file_id
  pslashindex=INSTRREV(file_id, "/") ' last slash => path <-/-> filename divider
  IF pslashindex=0 THEN ' only a filename / no path
    ' determine "local" path, and try to load
    file_id=LOCALEGETLOCALPATH(file_id)
    filename=file_id
    pslashindex=INSTRREV(file_id, "/")
  ELSE
    IF pslashindex=LEN(file_id) THEN ' no filename => <langcode>.str
      filename=file_id+lcode+".str"
      lcode=MID(lcode, 1, 2) ' lcode in next run
    ENDIF
  ENDIF
  ' #note: this code checks e.g. for
  ' ...locale/<(path/)file>
  ' ...locale/<language>/<language_region>/<file>
  ' ...locale/<language>/<file>
  ' ...locale/en/<file>
  ' and adds them in this order, but there shouldn't be the files 2-4 when file 1 exists
  ' the first file will be priorised anyway.

  WHILE filename<>""
    IF DFEXISTS(filename) THEN
      DFLOAD filename, fbuffer
      fbuffer+=CHR(13) ' last line must not be a string entry, linebreak marks string entries

      ' parsing file / shape it
      ' turn language section codes e.g. to \r[en_US\r
      ' add section IDs to string IDs => <section ID>.<string ID>
      ' every string entry gets "normalised" to \r[<section>.]<string ID>\x08<string>\r

      ' replace all line break codes with 13 (\r)
      ' replace all other control codes with Space (32)
      i=0
      l=LEN(fbuffer)
      DO
        c=fbuffer[i]
        IF c=13 THEN
          i+=1
          c=fbuffer[i]
          IF c=10 THEN ' DOS linebreak \r\n
            fbuffer=MID(fbuffer, 1, i)+MID(fbuffer, i+2)
            l-=1
          ENDIF
        ELSE
          IF c=10 THEN ' Unix linebreak \n
            fbuffer[i]=13
          ELSEIF c<32 THEN ' other ASCII control codes e.g. tabulator
            fbuffer[i]=32
          ENDIF
          i+=1
        ENDIF
      LOOP WHILE i<l
      IF fbuffer[i]<>13 THEN fbuffer+=CHR(13)
      IF fbuffer[0]<>13 THEN fbuffer=CHR(13)+fbuffer

      ' eliminating space lines and indentions and all spaces at the end of lines
      l=LEN(fbuffer)
      i=0
      DO
        c=fbuffer[i]
        IF c=13 THEN ' line break
          j=0
          i+=1
          WHILE i<l AND c=13
            c=fbuffer[i]
            i+=1
            j+=1
          WEND
          IF j>1 THEN ' replace sequence of \r with one \r
            IF i<l THEN
              i-=j
              fbuffer=MID(fbuffer, 1, i)+MID(fbuffer, i+j)
              j-=1
            ELSE ' it's the whole rest of the file...
              i-=j
              fbuffer=LEFT(fbuffer, i)
            ENDIF
            l-=j
          ELSEIF c=32 THEN ' space after line break
            i-=j
            j=0
            i+=1
            WHILE i<l AND c=32
              c=fbuffer[i]
              i+=1
              j+=1
            WEND
            IF j THEN
              ' remove all spaces at line start (indention)
              IF c=13 THEN j+=1 ' empty line out of spaces, additionally remove \r
              l-=j
              j+=1
              i-=j
              fbuffer=MID(fbuffer, 1, i)+MID(fbuffer, i+j)
            ENDIF
          ENDIF
        ELSEIF c=32 THEN ' space
          j=1
          i+=1
          WHILE i<l AND c=32
            c=fbuffer[i]
            i+=1
            j+=1
          WEND
          i-=j
          IF c=13 THEN ' sequence of spaces at the end of a line
            ' remove sequence of spaces
            fbuffer=MID(fbuffer, 1, i)+MID(fbuffer, i+j)
            l-=j
            i-=1
            l+=1
          ENDIF
        ENDIF
        i+=1
      LOOP WHILE i<l

      ' now turning the str file into \r[<lang.code>\r / \r<string id>\t<string>\r form
      i=1 ' first char is 13, keep it
      sectionid=""
      DO
        ' at this point, i points to the first char of a new line - fbuffer[i] - 'i+1' for MID etc.
        ' skip/remove empty lines
        j=i
        l=LEN(fbuffer)
        c=fbuffer[i]
        WHILE c=13 AND i<l
          i+=1
          c=fbuffer[i]
        WEND
        IF i=l THEN ' end of file, remove rest
          fbuffer=MID(fbuffer, 1, j) ' (keep previous \r)
          EXIT DO
        ELSEIF i>j THEN ' remove empty lines
          fbuffer=MID(fbuffer, 1, j)+MID(fbuffer, i+1)
          i=j
        ENDIF

        IF c=91 THEN ' '[' => section start (only valid on line start)
          j=i+2
          i+=1
          c=fbuffer[i]
          WHILE c=46 OR c=95 OR (c>=48 AND c<=57) OR (c>=65 AND c<=90) OR (c>=97 AND c<=122)
            ' valid ID char (letters+numbers & point & underscore)
            i+=1
            c=fbuffer[i]
          WEND

          IF c=93 THEN ' ']' => valid section code
            l=i-j+1
            sectionid=MID(fbuffer, j, l)
            ' i to next line
            WHILE c<>13 ' as long not a line break
              i+=1
              c=fbuffer[i]
            WEND
            IF (l=5 AND fbuffer[j+1]=95) OR (l=2) THEN ' #todo: AOT 3-letter lang. codes
              ' seems to be a language code ("xx_XX" or "xx")
              ' => "\r[<language code>\r"
              fbuffer=MID(fbuffer, 1, j-1)+sectionid+MID(fbuffer, i+1)
              sectionid=""
              i=INSTR(j+1, fbuffer, CHR(13)) ' i points to start of next line (behind \r)
            ELSE ' section name => remove line
              ' sectionid will be put before every following string ID until new section
              UCASE(sectionid, 0)
              fbuffer=MID(fbuffer, 1, j-2)+MID(fbuffer, i+2)
              i=j-2
            ENDIF
          ELSE ' not a valid section => remove line
            ' (can't be a valid string entry too)
            WHILE c<>13 ' as long not a line break
              i+=1
              c=fbuffer[i]
            WEND
            ' fbuffer[j] is 2 behind '\r['
            ' remove all between last \r including next \r
            fbuffer=MID(fbuffer, 1, j-2)+MID(fbuffer, i+2)
            i=j-2
          ENDIF
        ELSE ' not a section code
          ' search for string ID, ignore spaces
          WHILE c<>13 AND c<33 ' space chars except return
            i+=1
            c=fbuffer[i]
          WEND
          IF c=33 OR c=39 THEN ' comment char => skip line
            WHILE c<>13
              i+=1
              c=fbuffer[i]
            WEND
            fbuffer=MID(fbuffer, 1, j)+MID(fbuffer, i+2)
            i=j
          ELSE
            IF c=46 OR c=95 OR (c>=48 AND c<=57) OR (c>=65 AND c<=90) OR (c>=97 AND c<=122) THEN
              ' valid ID char (letters+numbers & point & underscore)
              ' determine (maybe-)string ID
              WHILE c=46 OR c=95 OR (c>=48 AND c<=57) OR (c>=65 AND c<=90) OR (c>=97 AND c<=122)
                i+=1
                c=fbuffer[i]
              WEND
              keyid=MID(fbuffer, j+1, i-j)
              UCASE(keyid, 0)
              IF sectionid="" THEN
                fbuffer=MID(fbuffer, 1, j)+keyid+CHR(8)+MID(fbuffer, i+1)
              ELSE
                fbuffer=MID(fbuffer, 1, j)+sectionid+"."+keyid+CHR(8)+MID(fbuffer, i+1)
              ENDIF
              i=INSTR(j, fbuffer, CHR(8))

              ' get next(first) quotation mark in line
              j=i
              c=fbuffer[i]
              WHILE c<>13 AND c<>34 AND c<>33 AND c<>39
                i+=1
                c=fbuffer[i]
              WEND
              IF c=13 THEN ' line end reached => test for multiline entry => search valid "END"
                IF i+5>LEN(fbuffer) THEN
                  ' end of file - invalid => remove last line
                  i=INSTRREV(fbuffer, CHR(13), i)
                  fbuffer=MID(fbuffer, i)
                  EXIT DO
                ENDIF
                ' remove rest of line (spaces+line break)
                fbuffer=MID(fbuffer, 1, j)+MID(fbuffer, i+2)
                ' valid 'END' line contains nothing else, only spaces are tolerated
                i=INSTR(i, fbuffer, CHR(13)+"END")+3
                WHILE i>3
                  j=i-4
                  c=fbuffer[i]
                  WHILE c<>13 AND c<33
                    c=fbuffer[i]
                    i+=1
                  WEND
                  IF c=13 THEN ' valid 'END' line
                    i+=1
                    fbuffer=MID(fbuffer, 1, j)+MID(fbuffer, i)
                    ' going backwards to code 8, replacing all line break codes with Space
                    i=j
                    WHILE c<>8
                      i-=1
                      c=fbuffer[i]
                      IF c=13 THEN
                        fbuffer[i]=32 ' Space
                        i-=1
                        c=fbuffer[i]
                      ENDIF
                    WEND
                    i=INSTR(i+2, fbuffer, CHR(13))
                    EXIT WHILE
                  ELSE ' error, not an 'END' line, get next candidate
                    i=INSTR(i, fbuffer, CHR(13)+"END")+3
                  ENDIF
                WEND

                IF i=3 THEN ' no valid 'END' found
                  ' remove (invalid) line and go on
                  j=INSTRREV(fbuffer, CHR(8))
                  i=INSTRREV(fbuffer, CHR(13), j)
                  j=INSTR(j, fbuffer, CHR(13))
                  fbuffer=MID(fbuffer, 1, i-1)+MID(fbuffer, j)
                ENDIF
              ELSEIF c=34 THEN ' quotation mark
                ' eliminate everything between [8] and/including ["]
                fbuffer=MID(fbuffer, 1, j)+MID(fbuffer, i+2)
                i=j

                ' get next quotation mark / string content
                c=fbuffer[i]
                WHILE c<>13 AND c<>34
                  i+=1
                  c=fbuffer[i]
                  ' backslash => masked char
                  IF c=92 THEN
                    i+=2
                    c=fbuffer[i]
                  ENDIF
                WEND
                IF c=13 THEN ' line end reached
                  ' remove (invalid) line
                  j=INSTRREV(fbuffer, CHR(13), j)
                  fbuffer=MID(fbuffer, 1, j)+MID(fbuffer, i+1)
                  i=j
                ELSE ' add string content, finish entry
                  fbuffer=MID(fbuffer, 1, i)+CHR(13)+MID(fbuffer, i+2)
                  i+=1
                ENDIF
              ELSE ' comment sign => illegal line, remove
                i=INSTRREV(fbuffer, CHR(13), i)
                j=INSTR(i+1, fbuffer, CHR(13))
                i-=1
                IF i THEN
                  fbuffer=MID(fbuffer, 1, i-1)+MID(fbuffer, j)
                ELSE
                  fbuffer=MID(fbuffer, j)
                ENDIF
              ENDIF
            ELSE ' invalid ID char => skip line
              ' i to next line
              WHILE c<>13 ' not a line break
                i+=1
                c=fbuffer[i]
              WEND
              fbuffer=MID(fbuffer, 1, j)+MID(fbuffer, i+2)
              i=j
            ENDIF
          ENDIF
        ENDIF
        IF i>=LEN(fbuffer) THEN EXIT DO
      LOOP
      localestrfile(index)+=fbuffer
    ENDIF
    IF lcode="" THEN EXIT WHILE ' all through

    IF LEN(lcode)=5 THEN ' <lang>_<region>
      filename=MID(file_id, 1, pslashindex)+MID(lcode, 1, 2)+"/"+lcode+MID(file_id, pslashindex)
      lcode=MID(lcode, 1, 2) ' lcode in next run
    ELSE
      IF pslashindex=LEN(file_id) THEN
        filename=MID(file_id, 1, pslashindex)+lcode+".str"
      ELSE
        filename=MID(file_id, 1, pslashindex)+lcode+MID(file_id, pslashindex)
      ENDIF
      IF lcode="en" THEN lcode="" ELSE lcode="en"
    ENDIF
  WEND
END SUB

SUB LOCALEUNLOADTABLE(BYVAL file_id AS STRING)
  DIM AS INTEGER index

  IF localecurlanguage="" THEN RETURN ' not initialised

  index=localestrfilecountsize
  DO
    index-=1
  LOOP WHILE index AND file_id<>localestrfilename(index)

  IF file_id=localestrfilename(index) THEN
    ' erase buffers
    localestrfilename(index)=""
    localestrfile(index)=""
  ENDIF

  IF index=localestrfilecountsize-1 THEN ' was last entry => reduce arrays
    localestrfilecountsize-=1
    ' also consider earlier emptied buffers
    WHILE localestrfilename(localestrfilecountsize)=""
      localestrfilecountsize-=1
    WEND
    REDIM PRESERVE localestrfilename(localestrfilecountsize)
    REDIM PRESERVE localestrfile(localestrfilecountsize)
  ENDIF
END SUB

FUNCTION LOCALEGETSTRING(BYVAL id AS STRING) AS STRING
  DIM AS INTEGER findex, i, j
  DIM AS STRING sstring, strcontent
  DIM AS UBYTE c
  DIM AS UBYTE ucode

  IF localecurlanguage="" THEN LOCALENEW ' forgot to initialise

  ' note: the stringtable data is "well-formed", so it is unnecessary to
  '       ensure additionally that INSTR() finds a linebreak etc.
  sstring=CHR(13)+id+CHR(8)
  UCASE(sstring, 0)
  findex=localestrfilecountsize
  WHILE findex
    findex-=1
    i=INSTR(localestrfile(findex), CHR(13)+"[")
    IF i THEN ' language code/section found
      ' #todo: ensure search in proper language section(s)
      ' only the first section of a distinct language is counting/valid,
      ' so you have only to find a string ID between its section start and end
      ' user-defined subsections are stored as <subsection>.<string ID>,
      ' any line starting with "[" is therfore a new language section
      i=INSTR(localestrfile(findex), sstring)
    ELSE ' whole file are strings in the same language
      i=INSTR(localestrfile(findex), sstring)
    ENDIF
    IF i THEN ' string found, extract/return
      j=INSTR(i, localestrfile(findex), CHR(8))
      j+=1
      i=INSTR(j, localestrfile(findex), CHR(13))
      ' string content (table)
      strcontent=MID(localestrfile(findex), j, i-j)
      ' translating masked chars
      i=1
      i=INSTR(i, strcontent, "\")
      WHILE i
        c=strcontent[i] AND 223
        ' some "escape codes" are replaced with control characters before returned
        SELECT CASE c
          CASE 69 ' E
            strcontent[i]=27
          CASE 78 ' N - newline Unix/DOS code "Line Feed", shouldn't be used...
            strcontent[i]=10
          CASE 82 ' R - return / Carriage Return
            strcontent[i]=13
          CASE 84 ' T - tabulator
            strcontent[i]=8
          CASE 85 ' U \u+XXXX for hexadecimal Unicode
            IF strcontent[i+2]=43 THEN
              ucode=VAL("&H"+MID(strcontent, i+3, 4))
              strcontent=MID(strcontent, 1, i)+UTF8(ucode)+MID(strcontent, i+7)
            ENDIF
          CASE 88 ' X \xXX for hexadecimal ASCII Code
            strcontent[i]=VAL("&H"+MID(strcontent, i+2, 2))
        END SELECT
        ' removing the '\'
        strcontent=MID(strcontent, 1, i-1)+MID(strcontent, i+1)
        i=INSTR(i, strcontent, "\")
      WEND
      RETURN strcontent ' success. all done
    ENDIF
  WEND

  RETURN "!ERR_NOT_FOUND:"+id+"!" ' failure
END FUNCTION

FUNCTION LOCALEGETLANGUAGELABEL(BYVAL language AS STRING) AS STRING
  DIM AS INTEGER i, j
  DIM AS STRING sstring, strcontent
  DIM AS UBYTE c
  DIM AS UBYTE ucode

  IF localecurlanguage="" THEN LOCALENEW ' forgot to initialise

  sstring=CHR(13)+"["+language+CHR(13)
  i=INSTR(localestrfile(0), sstring)

  sstring=CHR(13)+"IDS_LANGLABEL"+CHR(8)

  i=INSTR(i, localestrfile(0), sstring)
  IF i THEN ' string found, extract/return
    j=INSTR(i, localestrfile(0), CHR(8))
    j+=1
    i=INSTR(j, localestrfile(0), CHR(13))
    ' string content (table)
    strcontent=MID(localestrfile(0), j, i-j)
    ' translating masked chars
    i=1
    i=INSTR(i, strcontent, "\")
    WHILE i
      c=strcontent[i] AND 223
      ' some "escape codes" are replaced with control characters before returned
      SELECT CASE c
        CASE 69 ' E
          strcontent[i]=27
        CASE 78 ' N - newline Unix/DOS code "Line Feed", shouldn't be used...
          strcontent[i]=10
        CASE 82 ' R - return / Carriage Return
          strcontent[i]=13
        CASE 84 ' T - tabulator
          strcontent[i]=8
        CASE 85 ' U \u+XXXX for hexadecimal Unicode
          IF strcontent[i+2]=43 THEN
            ucode=VAL("&H"+MID(strcontent, i+3, 4))
            strcontent=MID(strcontent, 1, i)+UTF8(ucode)+MID(strcontent, i+7)
          ENDIF
        CASE 88 ' X \xXX for hexadecimal ASCII Code
          strcontent[i]=VAL("&H"+MID(strcontent, i+2, 2))
      END SELECT
      ' removing the '\'
      strcontent=MID(strcontent, 1, i-1)+MID(strcontent, i+1)
      i=INSTR(i, strcontent, "\")
    WEND
    RETURN strcontent ' success. all done
  ENDIF

  RETURN "!ERR_NOT_FOUND:IDS_LANGLABEL!" ' failure
END FUNCTION

FUNCTION LOCALEGETLANGUAGEPACK(BYVAL file_id AS STRING="", BYREF count AS INTEGER=0) AS STRING
  IF localecurlanguage="" THEN LOCALENEW ' forgot to initialise
  ' (files needn't beeing loaded for this check)

  '#todo:
  ' von der angegebenen STR-Datei alle existierenden Sprachversionen zurckgeben
  ' => ber Datei- und Ordnernamen locale/en.str locale/en/name.str
  '    & ber Inhalt /locale/main.str
  ' basiert nicht auf geladenen Dateien, eine "main.str" wird nach allen Sprachen durchsucht

  ' liste erzeugen: "en,en_US,en_UK,de,de_AT,de_DE,..."
  '  localelanguagepacklist
  '  localelanguagepackindex

  RETURN "" ' error
END FUNCTION

FUNCTION LOCALETIME(BYVAL unixtime AS LONGINT, BYVAL formatstr AS STRING="") AS STRING
  DIM AS INTEGER i, l
  DIM AS STRING result, sstr
  DIM AS UBYTE c
  DIM AS DATETIME dt

  IF localecurlanguage="" THEN LOCALENEW ' forgot to initialise
  IF formatstr="" THEN ' use default; short time (hh:mm)
    formatstr=LOCALEGETSTRING("IDS_SHORTTIME")
  ENDIF

  ' #todo: add support of different calendar systems than Gregorian
  ' default calendar system, others not yet supported
  IF LEFT(formatstr, 11)="!GREGORIAN:" THEN
    formatstr=MID(formatstr, 12)
  ENDIF

  dt=unixtime ' calculates all values from Unix time

  i=0
  l=LEN(formatstr)
  WHILE i<l
    c=formatstr[i]
    IF c<>37 THEN ' not percent sign %
      IF c=92 THEN ' Backslash \ => masking
        i+=1
        IF i>=l THEN
          result+="\"
          EXIT WHILE
        ENDIF
        c=formatstr[i]
        ' only "processing" \% \\ and \"
        IF c=37 OR c=92 OR c=34 THEN
          result+=CHR(c)
        ELSE
          result+="\"+CHR(c) ' copy 1:1, shouldn't be a problem
        ENDIF
      ELSE
        result+=CHR(c)
      ENDIF
    ELSE ' percent sign
      i+=1
      c=formatstr[i]
      IF c=95 THEN ' Underscore _
        ' => %_h, %_H, %_D, %_Dth, %_M
        sstr=MID(formatstr,i,3)
        SELECT CASE sstr
          CASE "%_h" ' hour 1-12
            c=(dt.hour MOD 12)
            IF c=0 THEN c=12
          CASE "%_H" ' hour 0-23
            c=dt.hour
          CASE "%_D" ' day of month
            c=dt.mday
            IF MID(formatstr,i,5)="%_Dth" THEN ' %_Dth
              IF c<10 THEN result+=" "
              result+=LOCALEORDINALNUMBER(c)
              i+=4
              CONTINUE WHILE
            ENDIF
          CASE "%_M" ' month
            c=dt.month
          CASE ELSE ' invalid, ignore, copy 1:1
            result+="%_"
            i+=1
            CONTINUE WHILE
        END SELECT
        IF c<10 THEN
          result+=" "+STR(c)
        ELSE
          result+=STR(c)
        ENDIF
        i+=1
      ELSEIF c=115 THEN ' s
        c=dt.second
        IF formatstr[i+1]=115 THEN ' %ss
          IF c<10 THEN result+="0"
          i+=1
        ENDIF ' ELSE %s
        result+=STR(c)
      ELSEIF c=109 THEN ' m
        sstr=MID(formatstr,i,6)
        IF sstr="%month" THEN ' %month
          c=dt.month
          result+=localegetstring("IDS_DATETIME_MONTH_"+STR(c))
          i+=4
        ELSE
          c=dt.minute
          IF formatstr[i+1]=109 THEN ' %mm
            IF c<10 THEN result+="0"
            i+=1
          ENDIF ' ELSE %m
          result+=STR(c)
        ENDIF
      ELSEIF c=104 THEN ' h
        c=dt.hour MOD 12
        IF c=0 THEN c=12
        IF formatstr[i+1]=104 THEN ' %hh
          IF c<10 THEN result+="0"
          i+=1
        ENDIF ' ELSE %h
        result+=STR(c)
      ELSEIF c=119 THEN ' w
        sstr=MID(formatstr,i,5)
        IF sstr="%wday" THEN ' %wday
          c=dt.wday+1
          result+=LOCALEGETSTRING("IDS_DATETIME_WEEKDAY_"+STR(c))
          i+=3
        ELSE
          result+="%w" ' invalid, copy 1:1
        ENDIF
      ELSEIF c=72 THEN ' H
        c=dt.hour
        IF formatstr[i+1]=72 THEN ' %HH
          IF c<10 THEN result+="0"
          i+=1
        ENDIF ' ELSE %H
        result+=STR(c)
      ELSEIF c=68 THEN ' D
        c=dt.mday
        sstr=MID(formatstr,i,4)
        IF sstr="%Dth" THEN ' %Dth
          result+=LOCALEORDINALNUMBER(c)
          i+=2
        ELSE
          IF formatstr[i+1]=68 THEN ' %DD
            IF c<10 THEN result+="0"
            i+=1
          ENDIF ' ELSE %D
          result+=STR(c)
        ENDIF
      ELSEIF c=77 THEN ' M
        sstr=MID(formatstr,i,6)
        IF sstr="%Month" THEN ' %Month
          sstr=LOCALEGETSTRING("IDS_DATETIME_MONTH_"+STR(dt.month))
          result+=UTF8UPPERCASE(sstr,1)
          i+=4
        ELSEIF LEFT(sstr,4)="%Mon" THEN ' %Mon
          result+=LOCALEGETSTRING("IDS_DATETIME_MON_"+STR(dt.month))
          i+=2
        ELSE
          i+=1
          c=formatstr[i]
          IF c=77 THEN ' %MM
            c=dt.month
            IF c<10 THEN
              result+="0"+STR(c)
            ELSE
              result+=STR(c)
            ENDIF
          ELSEIF c=110 THEN ' %Mn
            result+=LOCALEGETSTRING("IDS_DATETIME_MN_"+STR(dt.month))
          ELSE ' => %M
            result+=STR(dt.month)
            CONTINUE WHILE
          ENDIF
        ENDIF
      ELSEIF c=89 THEN ' Y
        ' %Y %YY %YYYY %Year
        sstr=MID(formatstr,i,5)
        IF sstr="%Year" THEN
          IF dt.year>0 THEN
            result+=STR(dt.year)
          ELSE
            result+=STR(ABS(dt.year)+1)
          ENDIF
        ELSEIF sstr="%YYYY" THEN
          IF dt.year>999 OR dt.year<-99 THEN
            result+=STR(dt.year)
          ELSE
            sstr=STR(dt.year)
            result+=STRING(4-LEN(sstr), "0")+sstr
          ENDIF
          i+=3
        ELSE
          IF (i+1)<l AND formatstr[i+1]=89 THEN ' %YY
            result+=STR(dt.year MOD 100)
            i+=1
          ELSE ' %Y
            result+=STR(dt.year)
          ENDIF
        ENDIF
      ELSEIF c=87 THEN ' W
        ' %Wd %Wdy %Wday
        c=dt.wday
      ELSEIF c=97 THEN ' a
        ' %ap (AM PM)
        c=dt.hour
        IF c<12 THEN
          result+=LOCALEGETSTRING("IDS_DATETIME_AM")
        ELSE
          result+=LOCALEGETSTRING("IDS_DATETIME_PM")
        ENDIF
        i+=1
      ELSEIF c=98 THEN ' b
        ' %ba (BC AD)
        IF dt.year<1 THEN
          result+=LOCALEGETSTRING("IDS_DATETIME_BC") ' before Christ
        ELSE
          result+=LOCALEGETSTRING("IDS_DATETIME_AD") ' anno Domini
        ENDIF
        i+=1
      ELSE
        result+="%"+CHR(c) ' invalid code, copy 1:1
      ENDIF
    ENDIF
    i+=1
  WEND

  '                          %s "0"-"59" second, %ss "00-59" second
  '                          %m 0-59 minute, %mm 00-59 minute
  '
  '                          %h 1-12 hour, %hh 01-12 hour, %H 0-23 hour, %HH 00-23 hour (%_h,%_H) %ap "AM"/"PM"
  '                          %D 1-31 day, %DD 01-31 day (%_D)
  '                          %Wd "Mo"-"Su" weekday, %Wdy "Mon"-"Sun"(3 chars => also "Mo."), %Wday "Monday"-"Sunday" (+%wday)
  '                          %Dth ordinal day number "1st"-"31th" (of course no "01st", "02nd", but %_Dth)
  '                          %M 1-12 month, %MM 01-12 (%_M), %Mn "Ja"-"De", %Mon "Jan"-"Dec", %Month "January"-"December"
  '                          %Y 1-nnnn year with 0, %YY 01-nn year, %YYYY -nnnn-nnnn year with year 0, %Year 1-nnnn BC/AD without 0
  '                          %ba "BC"/"AD" (important for %Year)
  '                          use \% to mask a percent sign, \\ and \" will be translated too but \ and " could be used without
  '                          escaping. Everything that is not a valid key or escape sequence will appear in the formatted string
  RETURN result
END FUNCTION

FUNCTION LOCALEDATE(BYVAL unixtime AS LONGINT, BYVAL formatstr AS STRING="") AS STRING
  IF localecurlanguage="" THEN LOCALENEW ' forgot to initialise
  IF formatstr="" THEN ' use default; short date (numeric year, month, day)
    formatstr=LOCALEGETSTRING("IDS_SHORTDATE")
  ENDIF

  RETURN LOCALETIME(unixtime, formatstr)
END FUNCTION

FUNCTION LOCALEORDINALNUMBER(BYVAL value AS INTEGER, BYVAL language AS STRING="") AS STRING
  DIM AS STRING result, sstr

  ' #todo: language="" => system language

  ' #todo: should only return strings for the actual language, fallback problem
  '        (some languages don't have a special "1st", just a "1." => suffix)
  sstr="IDS_ORDINAL_"+STR(value)
  result=LOCALEGETSTRING(sstr)

  IF LEFT(result,5)<>"!ERR_" THEN RETURN result

  result=LOCALEGETSTRING("IDS_ORDINALPREFIX")+STR(value)+LOCALEGETSTRING("IDS_ORDINALSUFFIX")

  RETURN result
END FUNCTION

FUNCTION LOCALENUMBER(BYVAL value AS DOUBLE, BYVAL digits AS INTEGER, BYVAL language AS STRING="") AS STRING
  RETURN ""
END FUNCTION

FUNCTION LOCALENUMBER(BYVAL value AS DOUBLE, BYVAL formatstr AS STRING="", BYVAL language AS STRING="") AS STRING
  RETURN ""
END FUNCTION
