'
' prj.bi
'
' FreeBASIC project info library
'
'
' PRJSET(STRING id, STRING title, STRING author, STRING version, STRING begin, STRING date)
'    sets the corresponding project values
'    dates have the format "YYYYMMDD hh:mm[+/-h]" where "hh" means the local hour,
'    UT+/-h("timezone"), e.g. "+2" for Central European Summer Time
'
' PRJGETID() returns the project application's identifier/name => executable's name
'
' PRJGETTITLE() returns the application's "full name" e.g for window title bar
'
' PRJGETAUTHOR() returns the program's author or the list of (most important) authors
'
' PRJGETVERSION() returns the program version in "A.B.C.D." format
'    "A" is the main version, a leading "a" marks an Alpha build, "b" a Beta version,
'    e.g. "a1.0.0.0" is 1.0 Alpha version
'    "B" is subversion/"milestone", "C" is for minor version steps,
'    "D" build/release number => to identify patches
'    a different main version is a different project, MyApp II should also
'    use its own files like "myapp_2.cfg"
'
' PRJGETBEGIN() returns the Unix timestamp of project begin
'
' PRJGETDATE() returns the Unix timestamp of current version(build)
'

#INCLUDE ONCE "utf8.bi"

DIM SHARED AS STRING prjid, prjtitle, prjauthor, prjversion
DIM SHARED AS SHORT prjversionvalue(5)
DIM SHARED AS INTEGER prjbegin, prjdate

FUNCTION PRJGETUNIXTIME(BYVAL tstring AS STRING) AS INTEGER
  DIM AS INTEGER t, y, d, v
  ' #note: imperfect Unix time calculation algorithm, only working because
  '        2000 was a leap year, maybe not 2038 safe (depending on INTEGER type)

  y=VAL(MID(tstring,1,4)) ' year
  IF y<1970 OR y>2099 THEN RETURN 0
  t=(y-1970)*31536000 ' year -> seconds
  
  v=VAL(MID(tstring,5,2)) ' month
  IF v<1 OR v>12 THEN RETURN 0
  IF v>1 THEN
    IF v>2 THEN 
      t+=INT((y-1968)/4)*86400 ' leap years -> seconds for Feb. 29th
      d=31+28 ' January+February(remaining days)
      d+=(v-3)*30 ' rest of months
      IF v<8 THEN
        d+=(v-2)/2 ' March+May
      ELSE
        d+=2 ' March+May
        d+=(v-7)/2
      ENDIF
      t+=d*86400
    ELSE
      t+=INT((y-1969)/4)*86400 ' leap years -> seconds for Feb. 29th
      t+=31*86400 ' January
    ENDIF
  ENDIF

  v=VAL(MID(tstring,7,2)) ' day
  IF v<1 OR v>31 THEN RETURN 0
  v-=1
  t+=v*86400
  v=VAL(MID(tstring,10,2)) ' hour
  IF v<0 OR v>23 THEN RETURN 0
  t+=v*3600
  v=VAL(MID(tstring,13,2)) ' minute
  t+=v*60
  t-=VAL(MID(tstring,15))*3600 ' "timezone"/offset

  RETURN t
END FUNCTION


SUB PRJSET(BYVAL id AS STRING="noname", BYVAL title AS STRING="Nameless Application", BYVAL author AS STRING="anonymous", BYVAL version AS STRING="a1.0.0.0", BYVAL begin AS STRING="", BYVAL latest_date AS STRING="")
  ' sets all PRJ values
  ' id = application ID => name of .EXE file and similar; short name of application
  ' title = application's title/full name, a.o.t. for the window bar
  ' author = author list, maybe software company, needn't be the full credits
  ' version = program version in form "[a/b]<main version>.<main subversion>.<sub-subversion>.<build number/counter>"
  ' begin = "YYYYMMDD HH:MM" time of project begin
  ' latest_date = "YYYYMMDD HH:MM" time of current build
  DIM AS INTEGER i, o, v

  IF id>"" THEN prjid=id
  IF title>"" THEN prjtitle=title
  IF author>"" THEN prjauthor=author
  IF version>"" THEN
    o=1
    IF MID(version,1,1)="a" THEN
      prjversion=UTF8(&H03B1) ' alpha
      prjversionvalue(4)=-2
      o+=1
    ELSEIF MID(version,1,1)="b" THEN
      prjversion=UTF8(&H03B2) ' beta
      o+=1
      prjversionvalue(4)=-1
    ELSE
      prjversionvalue(4)=0
    ENDIF

    i=0
    WHILE INSTR(o,version,".")
      prjversionvalue(i)=VAL(MID(version,o,INSTR(o,version,".")))
      o=INSTR(o,version,".")+1
      prjversion=prjversion+STR(prjversionvalue(i))+"."
      i+=1
    WEND
    prjversionvalue(i)=VAL(MID(version,o))
    prjversion=prjversion+STR(prjversionvalue(i))
  ENDIF
  IF begin>"" THEN
    prjbegin=PRJGETUNIXTIME(begin)
  ENDIF
  IF latest_date>"" THEN
    prjdate=PRJGETUNIXTIME(latest_date)
  ENDIF
END SUB


FUNCTION PRJGETID() AS STRING
  RETURN prjid
END FUNCTION


FUNCTION PRJGETTITLE() AS STRING
  RETURN prjtitle
END FUNCTION


FUNCTION PRJGETAUTHOR() AS STRING
  RETURN prjauthor
END FUNCTION


FUNCTION PRJGETVERSION() AS STRING
  RETURN prjversion
END FUNCTION


FUNCTION PRJGETBEGIN() AS INTEGER
  RETURN prjbegin
END FUNCTION


FUNCTION PRJGETDATE() AS INTEGER
  RETURN prjdate
END FUNCTION
